from typing import List


class DirectedAcyclicGraph:
    def __init__(self, steps: List):
        self.node_count = len(steps)
        self.adj_matrix = self.create_adj_matrix()
        self.nodes = steps

        name_to_node_map = {n.name: n for n in steps}
        for node in steps:
            dependencies = node.get_dependencies()
            for dep in dependencies:
                self.add_edge(steps.index(name_to_node_map[dep]), steps.index(node))

    def create_adj_matrix(self):
        result = [None] * self.node_count

        for i in range(self.node_count):
            result[i] = [False] * self.node_count

        return result

    def topological_sort(self):
        stack = []
        visited = [False] * len(self.nodes)

        for i in range(self.node_count):
            if not visited[i]:
                self.topological_sort_utils(i, visited, stack)

        result = []
        for i in stack:
            result.append(self.nodes[i])

        return result

    def add_edge(self, node, dependent_node):
        self.adj_matrix[dependent_node][node] = True

    def topological_sort_utils(self, v, visited, stack):
        visited[v] = True

        for index in range(self.node_count):
            if self.adj_matrix[v][index]:
                if not visited[index]:
                    self.topological_sort_utils(index, visited, stack)

        stack.append(v)



from convergence.convergence_service import ConvergenceService


def convergence_endpoint(router, url, method='GET', authorization=None, is_public_endpoint=False):
    if not is_public_endpoint and authorization is None:
        raise ValueError(f'An endpoint must either have authorization rule or be a public endpoint ({method}: {url})')
    ConvergenceService.register_endpoint_info(url, method, authorization, is_public_endpoint)

    def wrapper(func):
        if method.upper() == 'GET':
            router.get(url)(func)
        elif method.upper() == 'POST':
            router.post(url)(func)
        elif method.upper() == 'PUT':
            router.put(url)(func)
        elif method.upper() == 'PATCH':
            router.patch(url)(func)
        elif method.upper() == 'DELETE':
            router.delete(url)(func)

        return func

    return wrapper



from convergence.db_migrations.blueprint_formatter import BlueprintFormatter
from convergence.db_migrations.structures import TablesRelationshipBlueprint


class BaseTableRelationFormatter(BlueprintFormatter):
    def to_sql(self, blueprint: TablesRelationshipBlueprint):
        return f'ALTER TABLE {blueprint.table} ADD FOREIGN KEY ({blueprint.foreign_key}) REFERENCES {blueprint.primary_table}({blueprint.primary_key});'



import jwt
from jwt import DecodeError
from starlette.middleware.base import BaseHTTPMiddleware
from fastapi import Request
from starlette.responses import JSONResponse

from convergence.dto.base_dto import ApiResponse
from convergence.helpers import errors
from convergence.helpers.controller_helpers import convert_object_to_dict, update_response_with_request_ids_for_filters
from convergence.internal.observability.request_log import RequestLog
from convergence.internal.observability.serializer_config import LogSerializerConfig


class AuthorizationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, **kwargs):
        super().__init__(app)
        self.service = kwargs['service']
        self.signing_key = self.service.get_configuration('security.authentication.secret')
        self.signing_key = self.signing_key.replace('\\n', '\n')
        self.log_serializer = LogSerializerConfig.create(self.service)
        self.request_type = self.service.get_configuration('observability.request_id_prefix')

    async def dispatch(self, request: Request, call_next):
        path = request.scope.get('path')
        method = request.scope.get('method')
        endpoint_info, path_matched = self.service.get_endpoint_info(path, method)

        if endpoint_info is None:
            if path_matched:
                response = self.__not_allowed_method_response(path, request)
            else:
                response = self.__not_found_response(path, request)
        else:
            auth_header = request.headers.get('Authorization')

            payload = None
            if auth_header is not None:
                err_code, message, payload = self.__is_valid_authorization_token(auth_header)
                if err_code is not None:
                    return self.__invalid_authorization_header(err_code, message, request)

            if endpoint_info.authorization is None or self.is_authorized(endpoint_info, request, payload):
                response = await call_next(request)
            else:
                response = self.__unauthorized_response(path, request)

        return response

    def is_authorized(self, endpoint_info, request, token_payload):
        acl = self.service.acl
        return acl.is_authorized(endpoint_info.authorization, request, token_payload)

    def __is_valid_authorization_token(self, auth_header):
        err_code = None
        err_message = None

        bearer_length = len('Bearer ')
        if auth_header[0:bearer_length] != 'Bearer ':
            err_code = errors.INVALID_AUTHORIZATION_TOKEN
            err_message = 'Service expects an authorization token in Bearer format.'
            return err_code, err_message, None

        try:
            auth_header = auth_header[bearer_length:]
            result = jwt.decode(auth_header, key=self.signing_key, algorithms="ES512")
            return None, None, result
        except DecodeError as ex:  # noqa
            err_code = errors.INVALID_AUTHORIZATION_TOKEN
            err_message = 'Service expects a valid JWT token.'

            return err_code, err_message, None

    def __invalid_authorization_header(self, err_code, message, http_request):
        request_log = RequestLog.initialize(self.request_type, http_request, self.service)

        status_code = 401

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = message
        response.header.code = err_code
        response.header.body_type = 'failure_info'

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    def __unauthorized_response(self, path, http_request):
        request_log = RequestLog.initialize(self.request_type, http_request, self.service)

        status_code = 403

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = f'The authorization token is invalid for path {path}'
        response.header.code = errors.INVALID_AUTHORIZATION_TOKEN
        response.header.body_type = 'failure_info'

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    def __not_found_response(self, path, http_request):
        request_log = RequestLog.initialize(self.request_type, http_request, self.service)

        status_code = 404

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = f'Unable to find resource at path {path}'
        response.header.code = errors.API_RESOURCE_NOT_FOUND
        response.header.body_type = 'failure_info'

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    def __not_allowed_method_response(self, path, http_request):
        request_log = RequestLog.initialize(self.request_type, http_request, self.service)

        status_code = 405

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = f'Unable to find resource at path {http_request.method.upper()} {path}'
        response.header.code = errors.API_METHOD_NOT_ALLOWED
        response.header.body_type = 'failure_info'

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)



import uuid


class ServiceAuthorityDeclaration:
    def __init__(self, id, authority, display_name, tier):
        self.uuid = uuid.UUID(id)
        self.authority = authority
        self.display_name = display_name
        self.tier = tier



import abc
import argparse
import sys
from typing import List

import uvicorn
from fastapi import FastAPI
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from convergence.internal.observability.service_stdout_redirection import ConsoleFileDualWriter
from convergence.internal.services.authentication_service import AuthenticationMicroService
from convergence.internal.services.infrastructure_service import InfrastructureMicroService
from convergence.internal.utils.service_configuration_loader import load_service_configuration
from convergence.metadata.service_authority import ServiceAuthorityDeclaration
import convergence.constants as constants
from starlette.middleware.cors import CORSMiddleware

from convergence.security.access_control_layer import AccessControlLayer
from convergence.security.authorization_filter import AuthorizationMiddleware
from convergence.security.gateway_header_validation_filter import GatewayHeaderValidationMiddleware
from convergence.security.input_validation_filter import InputValidationMiddleware


def create_cli_argument_parser():
    parser = argparse.ArgumentParser(description='Convergence Service')

    parser.add_argument('--profile',
                        type=str,
                        required=False,
                        help='Specify a profile to override the values in the default one.')

    return parser


def get_override_profile_file_path(args):
    if args.profile is not None:
        return f'./configurations/application-{args.profile}.yaml'

    return None


class ServiceState:
    def __init__(self):
        self.status = 'initializing'


class ConvergenceEndpointInfo:
    def __init__(self, url, method, authorization, is_public_endpoint):
        self.url = url
        self.method = method
        self.authorization = authorization
        self.is_public_endpoint = is_public_endpoint


def pad(value, length):
    return value + ''.join([' '] * (length - len(value)))


class ConvergenceService(abc.ABC):
    _instance = None
    __endpoints_info: List[ConvergenceEndpointInfo] = []

    def __init__(self, default_config_path=None, profile_config_path=None):
        if ConvergenceService._instance is not None:
            raise ValueError('There can only be one instance/extension of ConvergenceService.')

        ConvergenceService._instance = self
        self._configuration = load_service_configuration(default_config_path, profile_config_path)
        self.service_state = ServiceState()
        self._db_migration_module = 'db_migrations'

        kwargs = {}
        mode = self.get_configuration('application.mode')
        if mode == 'production':
            kwargs['openapi_url'] = None
        self.app = FastAPI(**kwargs)
        self.acl = AccessControlLayer()

    @staticmethod
    def instance():
        return ConvergenceService._instance

    def include_router(self, router):
        self.app.include_router(router)

    def start(self):
        info = self.get_service_info()
        print('Launching service with info:')
        print(f'   Name: {info.SERVICE_NAME}')
        print(f'   Version: {info.SERVICE_VERSION}')
        print(f'   Hash: {info.SERVICE_VERSION_HASH}')
        print('', flush=True)

        uvicorn.run(self.app, host="0.0.0.0", port=self.get_configuration('server.port'))
        if self.get_configuration('observability.enable_stout_redirection'):
            sys.stdout.flush()

    def get_configuration(self, path, default_value=None):
        try:
            components = path.split('.')
            result = self._configuration

            for c in components:
                result = result[c]

            return result
        except BaseException as e:  # noqa
            if default_value is None:
                raise e
            return default_value

    def initialize(self):
        self.__initialize_redirect_log_to_file()
        self.__print_figlet()
        self.__print_server_port()
        self.service_state.status = 'initializing_db'
        self.__initialize_orm()
        self.__migrate_database()
        self.service_state.status = 'db_initialized'
        self.__save_service_authority()
        self.service_state.status = 'healthy'
        self.__initialize_cors()
        self.__initialize_service_filters()

    def __initialize_orm(self):
        import convergence.database_orm as orm
        db_type = self.get_configuration('database.type')
        if db_type == 'postgres':
            user = self.get_configuration('database.username')
            password = self.get_configuration('database.password')
            host = self.get_configuration('database.host')
            db = self.get_configuration('database.name')
            port = self.get_configuration('database.port')
            orm.SQLALCHEMY_DATABASE_URL = f"postgresql://{user}:{password}@{host}:{port}/{db}"
            orm.DB_ENGINE = create_engine(orm.SQLALCHEMY_DATABASE_URL)

            orm.DBSessionLocal = sessionmaker(bind=orm.DB_ENGINE)
        else:
            raise ValueError('Unsupported database type')

    def __migrate_database(self):
        from convergence.internal.db.database_migrations import (
            load_database_migrations,
            get_list_of_migrations,
            initialize_formatters,
            create_connection_to_db,
            apply_database_migration,
            sort_migrations
        )

        migrations = load_database_migrations(self._db_migration_module)
        migrations = sort_migrations(migrations)
        formatters = initialize_formatters(self)

        with create_connection_to_db(self) as connection:
            print('Starting to apply database migrations:')
            applied_database_migrations = get_list_of_migrations(self, connection)

            error = None
            failed_migration = None
            for migration in migrations:
                if migration.name not in applied_database_migrations:
                    if failed_migration is not None:
                        print(f'   - {pad(migration.name + ":", 50)} [ SKIPPED ]')
                    else:
                        migration_error = apply_database_migration(self, migration, connection, formatters)
                        if migration_error is not None:
                            failed_migration = migration.name
                            error = migration_error
                            print(f'   - {pad(migration.name + ":", 50)} [ FAILED ]')
                        else:
                            print(f'   - {pad(migration.name + ":", 50)} [ SUCCESS ]')
                else:
                    print(f'   - {pad(migration.name + ":", 50)} [ ALREADY_APPLIED ]')

            if error is not None:
                raise ValueError(error)

    def __save_service_authority(self):
        self.service_state.status = "initializing_authorities"
        mode = self.get_configuration('application.mode')

        service_authorities_handler = self.get_service_authorities_handler()
        if str(service_authorities_handler).startswith('<class '):
            if mode == 'production':
                self.__register_class_service_authority(service_authorities_handler)
            else:
                print('Service is running in non-production mode, skipping the authorities initialization:')
                authorities = self.__get_declared_authorities(service_authorities_handler)
                if len(authorities) == 0:
                    print('   -> Service doesn\'t declares any authority')
                else:
                    for auth in authorities:
                        tier = pad(f'[Tier {auth.tier}]', 10)
                        print(f'   - {tier} {auth.authority}')
        elif str(service_authorities_handler).startswith('<method '):
            service_authorities_handler()

        self.service_state.status = "authorities_initialized"

    @abc.abstractmethod
    def get_service_authorities_handler(self):
        pass

    @abc.abstractmethod
    def get_service_info(self):
        pass

    def __register_class_service_authority(self, service_authorities_handler):
        authorities = self.__get_declared_authorities(service_authorities_handler)

        if len(authorities) > 0:
            print('Service is running in production mode, starting to initialize authorities:')
            infrastructure_service = InfrastructureMicroService(self)
            authentication_service_url = infrastructure_service.get_service_url('authentication-service')
            print(f'   -> Authentication Service URL: {authentication_service_url}')
            authentication_service = AuthenticationMicroService(self, authentication_service_url)

            any_failed = False
            for authority in authorities:
                if not self.__register_single_service_authority(authority, authentication_service):
                    print(f'   * {pad(authority.authority + ":", 50)} [ FAILED INITIALIZATION ]')
                    any_failed = True
                else:
                    print(f'   * {pad(authority.authority + ":", 50)} [ INITIALIZED ]')

            if any_failed:
                raise ValueError('Service was unable to initialize the necessary authorities.')
        else:
            print('Service is running in production mode, but doesn\'t declare any authority')

    def __register_single_service_authority(self, authority, authentication_service):
        request = {
            'uuid': str(authority.uuid),
            'authority': authority.authority,
            'display_name': authority.display_name,
            'tier': authority.tier,
        }

        response = authentication_service.register_service_authority(request)
        return response['header']['status_code'] == 200

    def __get_declared_authorities(self, service_authorities_handler):
        authorities = []
        attributes = dir(service_authorities_handler)
        for attribute in attributes:
            value = getattr(service_authorities_handler, attribute)
            if isinstance(value, ServiceAuthorityDeclaration):
                authorities.append(value)
        return authorities

    def __print_figlet(self):
        print('   ' + """
   ______                                                    
  / ____/___  ____ _   _____  _________ ____  ____  ________ 
 / /   / __ \/ __ \ | / / _ \/ ___/ __ `/ _ \/ __ \/ ___/ _ \\
/ /___/ /_/ / / / / |/ /  __/ /  / /_/ /  __/ / / / /__/  __/
\____/\____/_/ /_/|___/\___/_/   \__, /\___/_/ /_/\___/\___/ 
                                /____/
        """.strip())  # noqa
        print(' :: Convergence Platform ::')
        print(f'     Version: {constants.VERSION}')
        print(f'     Hash: {constants.VERSION_HASH}')
        print(f'     Build Date: {constants.BUILD_DATE}')
        print('')

    def __print_server_port(self):
        print(f'Server will run on port {self.get_configuration("server.port")}')

    def __initialize_cors(self):
        is_enabled = self.get_configuration('cors.enabled', default_value=False)
        if is_enabled:
            print('Initializing CORS headers for the service.')
            self.app.add_middleware(
                CORSMiddleware,
                allow_origins=self.__to_list(self.get_configuration('cors.allow_origins')),
                allow_credentials=self.get_configuration('cors.allow_credentials'),
                allow_methods=self.__to_list(self.get_configuration('cors.allow_methods')),
                allow_headers=self.__to_list(self.get_configuration('cors.allow_headers')),
            )
        else:
            print('CORS header initialization is disabled for this service.')

    def __to_list(self, value):
        if not isinstance(value, list):
            value = [value]

        return value

    @staticmethod
    def register_endpoint_info(url, method, authorization, is_public_endpoint):
        ConvergenceService.__endpoints_info.append(ConvergenceEndpointInfo(url, method.upper(), authorization, is_public_endpoint))

    @staticmethod
    def get_endpoint_info(url: str, method: str):
        result = None
        method = method.upper()
        path_matched = False

        for info in ConvergenceService.__endpoints_info:
            url_matched = ConvergenceService.__match_url_to_endpoint(url, info.url)
            if url_matched:
                path_matched = True
            if info.method == method and url_matched:
                result = info
                break

        return result, path_matched

    @staticmethod
    def __match_url_to_endpoint(url, ep_url):
        if url == ep_url:
            return True

        comps_url = url.split('/')
        comps_ep_url = ep_url.split('/')

        if len(comps_url) == len(comps_ep_url):
            for a, b in zip(comps_url, comps_ep_url):
                if (b.startswith('{') and b.endswith('}')) or a == b:
                    continue
                return False

            return True
        else:
            return False

    def __initialize_service_filters(self):
        self.app.add_middleware(GatewayHeaderValidationMiddleware, service=self)
        self.app.add_middleware(AuthorizationMiddleware, service=self)
        self.app.add_middleware(InputValidationMiddleware, service=self)

    def __initialize_redirect_log_to_file(self):
        if self.get_configuration('observability.enable_stout_redirection'):
            dual_writer = ConsoleFileDualWriter(self, sys.stdout)
            sys.stdout = dual_writer
            sys.stderr = dual_writer



from convergence.internal.services.base_internal_service import BaseExternalService


class InfrastructureMicroService(BaseExternalService):
    def __init__(self, service):
        super().__init__(service, service.get_configuration('application.discovery_server_url'), service_jwt=True)

    def get_service_connection_details(self, service_name):
        request = {
            'service_name': service_name
        }

        return self.post_request('/infrastructure/get-service-info', request)



from typing import List

from convergence.dto.base_dto import ApiResponseBody


class BadFieldRequest:
    def __init__(self):
        self.location = ''
        self.field = ''
        self.error_messages = []


class BadRequestDetails(ApiResponseBody):
    def __init__(self):
        self.errors: List[BadFieldRequest] = []

    def get_response_body_type(self) -> str:
        return 'request_error_info'



import traceback

from convergence.dto.base_dto import FailureInfoDTO, create_api_response
import convergence.helpers.errors as errors
from convergence.internal.observability.request_log import RequestLog
from convergence.internal.observability.serializer_config import LogSerializerConfig

LOG_SERIALIZER = None


class ManagedApiError(BaseException):
    def __init__(self, status_code, code, message):
        self.status_code = status_code
        self.code = code
        self.message = message


def update_response_with_request_ids_for_filters(request_log: RequestLog, response, http_request):
    response.header.request_id = __remove_request_id_prefix(request_log.request_identifier)
    response.header.parent_request_id = request_log.parent_request_identifier


def __remove_request_id_prefix(request_identifier):
    if '_' in request_identifier:
        request_identifier = request_identifier[request_identifier.rindex('_') + 1:]
    return request_identifier


def run_api_method(request_log: RequestLog, service, http_request, http_response, func):
    global LOG_SERIALIZER
    if LOG_SERIALIZER is None:
        LOG_SERIALIZER = LogSerializerConfig.create(service)

    api_response = None
    try:
        api_response = create_api_response(func(), http_request)
    except BaseException as ex:
        tb = traceback.format_exc()
        request_log.exception(ex, tb)
        api_response = create_internal_error_response(service, http_request, errors.API_INTERNAL_ERROR, ex)

    http_response.status_code = api_response.header.status_code

    api_response.header.request_id = __remove_request_id_prefix(request_log.request_identifier)
    api_response.header.parent_request_id = request_log.parent_request_identifier

    request_log.finish(api_response)
    LOG_SERIALIZER.save(request_log)
    return api_response


def create_internal_error_response(service, http_request, code: str, ex: BaseException):
    failure = FailureInfoDTO()
    message = str(ex)

    if isinstance(ex, ManagedApiError):
        failure.code = ex.code
        failure.message = ex.message
        failure.status_code = ex.status_code

        return create_api_response(failure, http_request)
    elif service.get_configuration('application.mode') == 'production':
        message = "An unexpected error happened during API execution"

    failure.code = code
    failure.message = message
    failure.status_code = 500

    return create_api_response(failure, http_request)


def convert_object_to_dict(obj):
    if isinstance(obj, dict):
        return {str(key): convert_object_to_dict(item) for key, item in obj.items()}
    elif obj is None:
        return None
    elif isinstance(obj, int) or isinstance(obj, float):
        return obj
    elif not hasattr(obj, "__dict__"):
        return str(obj)

    result = {}
    skip_field = '_SKIP_LOGGING_FIELDS'
    secret_fields = getattr(obj, skip_field) if hasattr(obj, skip_field) else []
    for key, val in obj.__dict__.items():
        if key.startswith("_"):
            continue

        if isinstance(val, list):
            element = []
            for item in val:
                element.append(convert_object_to_dict(item))
        elif isinstance(val, set) or isinstance(val, tuple):
            element = []
            val = list(val)
            for item in val:
                element.append(convert_object_to_dict(item))
        else:
            element = convert_object_to_dict(val)
        result[key] = element

    for field in secret_fields:
        result[field] = '***********'

    return result



import threading
import time
import uuid

from fastapi import Request

from convergence.constants import (HEADER_REQUEST_ID,
                                   HEADER_PARENT_REQUEST_ID,
                                   HEADER_CALLER_SERVICE,
                                   HEADER_CALLER_SERVICE_VERSION,
                                   HEADER_CALLER_SERVICE_HASH)


class RequestLog:
    def __init__(self):
        self.request_identifier = ''
        self.parent_request_identifier = ''
        self.caller_service = ''
        self.receiver_service = ''
        self.start_timestamp = 0
        self.end_timestamp = 0
        self.headers = {}
        self.url = ''
        self.parameters = []
        self.log_entries = []
        self.response = None

    @staticmethod
    def initialize(request_type, http_request: Request, service, *parameters):
        result = RequestLog()

        result.start_timestamp = int(time.time() * 1000)
        result.headers = RequestLog.__load_request_headers(http_request)
        result.request_identifier, result.parent_request_identifier = RequestLog.__get_request_ids(http_request,
                                                                                                   request_type)
        result.caller_service = RequestLog.__load_caller_service(result.headers)
        result.receiver_service = RequestLog.__load_current_service(service)
        result.url = RequestLog.__get_request_uri(http_request)
        result.parameters = parameters

        RequestLog.__remove_header(result.headers, HEADER_REQUEST_ID)
        RequestLog.__remove_header(result.headers, HEADER_PARENT_REQUEST_ID)
        RequestLog.__remove_header(result.headers, HEADER_CALLER_SERVICE)
        RequestLog.__remove_header(result.headers, HEADER_CALLER_SERVICE_VERSION)
        RequestLog.__remove_header(result.headers, HEADER_CALLER_SERVICE_HASH)
        RequestLog.__remove_jwt_signature_for_logging(result.headers)

        return result

    def finish(self, response):
        self.response = response
        self.end_timestamp = int(time.time() * 1000)

    def info(self, message, *args, **kwargs):
        self.__add_log_entry('info', message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.__add_log_entry('error', message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.__add_log_entry('warning', message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        self.__add_log_entry('debug', message, *args, **kwargs)

    def exception(self, ex, trace_back):
        entry = LogEntry()
        entry.timestamp = int(time.time() * 1000)
        entry.level = 'exception'
        entry.message = str(ex)
        entry.arguments = None
        entry.named_arguments = {
            'format': 'python: trace_back',
            'stack_trace': trace_back
        }
        entry.type = "exception_entry"
        entry.thread_id = threading.current_thread().native_id

        self.log_entries.append(entry)

    def __add_log_entry(self, level, message, *args, **kwargs):
        entry = LogEntry()
        entry.timestamp = int(time.time() * 1000)
        entry.level = level
        entry.message = message
        entry.arguments = args
        entry.named_arguments = kwargs
        entry.type = "log_entry"
        entry.thread_id = threading.current_thread().native_id

        self.log_entries.append(entry)

    @classmethod
    def __load_request_headers(cls, http_request: Request):
        names = http_request.headers.keys()
        return {header: http_request.headers[header] for header in names}

    @classmethod
    def __load_caller_service(cls, headers):
        result = LogEntryServiceInfo()

        result.name = headers.get(HEADER_CALLER_SERVICE.lower())
        result.version = headers.get(HEADER_CALLER_SERVICE_VERSION.lower())
        result.hash = headers.get(HEADER_CALLER_SERVICE_HASH.lower())

        return result

    @classmethod
    def __load_current_service(cls, service):
        result = LogEntryServiceInfo()

        info = service.get_service_info()
        result.name = info.SERVICE_NAME
        result.version = info.SERVICE_VERSION
        result.hash = info.SERVICE_VERSION_HASH

        return result

    @classmethod
    def __get_request_uri(cls, http_request: Request):
        url = http_request.url.path[len(http_request.base_url.path):]
        if not url.startswith('/'):
            url = '/' + url

        return url

    @staticmethod
    def __get_request_ids(http_request, request_type):
        request_id = http_request.headers.get(HEADER_REQUEST_ID)
        if request_id is None:
            request_id = str(uuid.uuid4())

        request_id = request_type.lower() + '_' + request_id

        parent_request_id = http_request.headers.get(HEADER_PARENT_REQUEST_ID)

        return request_id, parent_request_id

    @classmethod
    def __remove_header(cls, headers, header):
        if header in headers:
            headers.pop(header)

        if header.lower() in headers:
            headers.pop(header.lower())

    @classmethod
    def __remove_jwt_signature_for_logging(cls, headers):
        header = None
        if 'authorization' in headers:
            header = 'authorization'
        if 'Authorization' in headers:
            header = 'Authorization'

        if header is not None:
            value = headers[header]
            if '.' in value:
                value = value[0:value.rindex('.') + 1] + '***********'
            headers[header] = value


class LogEntryServiceInfo:
    def __init__(self):
        self.name = ''
        self.version = ''
        self.hash = ''


class LogEntry:
    def __init__(self):
        self.timestamp = 0
        self.level = ''
        self.message = ''
        self.type = ''
        self.arguments = None
        self.named_arguments = None
        self.thread_id = ''



import datetime
import uuid

from convergence.db_migrations.blueprint_formatter import BlueprintFormatter


class BaseDatabaseSeedFormatter(BlueprintFormatter):
    def to_sql(self, blueprint):
        result = []

        for seed in blueprint:
            field_list = []
            values_list = []

            sep = ""
            for field in seed.fields:
                field_list.append(sep)
                values_list.append(sep)

                field_list.append(field)
                values_list.append(self.format_value(seed.fields[field]))
                sep = ", "

            result.append("INSERT INTO ")
            result.append(''.join(seed.table_name))
            result.append("(")
            result.append(''.join(field_list))
            result.append(") VALUES(")
            result.append(''.join(values_list))
            result.append("); \n\n")

        result = ''.join(result)
        return result

    def format_value(self, o):
        result = None

        if o is None:
            result = "NULL"
        elif isinstance(o, uuid.UUID):
            result = f"'{str(o)}'"
        elif isinstance(o, bool):
            result = 'true' if o else 'false'
        elif isinstance(o, datetime.datetime):
            result = o.strftime("%Y-%m-%d %H:%M:%S")
            result = f"'{result}'"
        elif isinstance(o, int) or isinstance(o, float):
            result = str(o)
        elif isinstance(o, str):
            lower = o.lower()
            forbidden = ['--', "'", 'create', 'drop', 'alter', 'insert', 'grant']
            for s in forbidden:
                if s in lower:
                    raise ValueError("This could be an SQL injection.")

            result = "'" + o + "'"
        else:
            raise ValueError()

        result = ''.join(result)
        return result



from convergence.dto.service_status import ServiceEndpointDTO


def matches_any(internal_endpoint_urls, path):
    return path in internal_endpoint_urls


def load_service_urls(service, internal_endpoint_urls):
    internal_endpoints = []
    public_endpoints = []
    for route in service.app.routes:
        if route.include_in_schema:
            endpoint_list = internal_endpoints if matches_any(internal_endpoint_urls, route.path) else public_endpoints
            ep = ServiceEndpointDTO()
            ep.url = route.path
            ep.methods = route.methods

            endpoint_list.append(ep)

    return internal_endpoints, public_endpoints



from convergence.db_migrations.blueprint_formatter import BlueprintFormatter
from convergence.db_migrations.structures import TableIndexBlueprint
from convergence.db_migrations.structures import TableBlueprint, TableColumnBlueprint


class BaseTableBlueprintFormatter(BlueprintFormatter):
    def to_sql(self, blueprint: TableBlueprint):
        result = []

        check_existence = ""
        if blueprint.check_existence:
            check_existence = "IF NOT EXISTS "

        result.append("CREATE TABLE ")
        result.append(check_existence)
        result.append(blueprint.name)
        result.append(" \n(")

        constraint_count = 0
        for column in blueprint.columns:
            if column.is_unique:
                constraint_count += 1

        index = 0
        for column in blueprint.columns:
            index += 1
            is_last_statement = index == len(blueprint.columns) and constraint_count == 0
            result.append("\n    ")
            result.append(self.get_column_definition(column))
            if not is_last_statement:
                result.append(",")

        index = 0
        for column in blueprint.columns:
            if column.is_unique:
                if index == 0:
                    result.append("\n")
                index += 1
                is_last_statement = index == constraint_count
                result.append(f"\n    CONSTRAINT {blueprint.name}_{column.name}_unique_index UNIQUE ({column.name})")

                if not is_last_statement:
                    result.append(",")

        result.append("\n)")

        if len(blueprint.indices) > 0:
            for indexBlueprint in blueprint.indices:
                result.append("\n\n")
                result.append(self.format_table_index(blueprint, indexBlueprint))
                result.append("")

        result = ''.join(result)
        return result

    def format_table_index(self, blueprint: TableBlueprint, index_blueprint: TableIndexBlueprint):
        table = blueprint.name
        fields_comma = []
        fields_underscore = []

        sc = ""
        su = ""

        for col in index_blueprint.columns:
            fields_comma.append(sc)
            fields_comma.append(col)
            fields_underscore.append(su)
            fields_underscore.append(col)
            sc = ", "
            su = "_"

        fields_underscore = ''.join(fields_underscore)
        fields_comma = ''.join(fields_comma)
        return f"CREATE INDEX {table}_{fields_underscore}_index ON {table} USING {index_blueprint.type}({fields_comma})"

    def get_column_definition(self, column: TableColumnBlueprint):
        result = []

        result.append(self.prepare_column_name_for_creation(column))
        result.append(' ')
        result.append(self.to_sql_type(column.type))
        result.append(' ')

        if column.is_primary_key:
            result.append("PRIMARY KEY ")

        if column.allow_null:
            result.append("NULL")
        else:
            result.append("NOT NULL")

        if column.default_value is not None:
            result.append(" DEFAULT ")
            result.append(column.default_value)

        result = ''.join(result)
        return result

    def prepare_column_name_for_creation(self, column: TableColumnBlueprint):
        return column.name

    def to_sql_type(self, type: str):
        result = ""

        if type == "String":
            result = "text"
        elif type.startswith("String[") and type.endswith("]"):
            length = type[7:-1]
            result = f"varchar({length})"
        elif type == "bool" or type == "boolean":
            result = "boolean"
        elif type.lower() == "json":
            result = "json"
        elif type == "int":
            result = "int"
        elif type == "date":
            result = "date"
        elif type == "double":
            result = "float8"
        elif type == "float":
            result = "float4"
        elif type == "timestamp":
            result = "timestamp"
        elif type.lower() == "uuid":
            result = "uuid"

        return result



from typing import List, Dict
from convergence.dto.base_dto import ApiResponseBody


class ServiceEndpointDTO:
    def __init__(self):
        self.url = ''
        self.methods: List[str] = []


class ServiceStatusDTO(ApiResponseBody):
    def __init__(self):
        self.service_name = ''
        self.version_hash = ''
        self.version = ''
        self.status = ''
        self.internal_endpoints: List[ServiceEndpointDTO] = []
        self.public_endpoints: List[ServiceEndpointDTO] = []
        self.extra: Dict[str, str] = {}

    def get_response_body_type(self) -> str:
        return 'service_status'



import time
from datetime import timezone, datetime, timedelta

import jwt
import requests


class BaseExternalService:
    def __init__(self, service, url, jwt_authority=None, service_jwt=False):
        self.url = url
        self.service = service
        self.jwt = self.__create_internal_jwt(service, jwt_authority, service_jwt)

    def __create_internal_jwt(self, service, authorities, service_jwt):
        service_info = service.get_service_info()
        signing_key = service.get_configuration('security.authentication.secret').replace('\\n', '\n')

        payload = {
            "sub": service_info.SERVICE_NAME,
            "iss": service_info.SERVICE_NAME,
            "exp": datetime.now(tz=timezone.utc) + timedelta(minutes=1),
        }

        if service_jwt:
            payload['is_inter_service_call'] = True

        if authorities is not None:
            authorities = [] if authorities is None else authorities
            authorities = [authorities] if not isinstance(authorities, list) else authorities
            payload['authorities'] = authorities

        return jwt.encode(payload, signing_key, algorithm="ES512")

    def get_service_url(self, service_name):
        from convergence.internal.services.infrastructure_service import InfrastructureMicroService
        infra_service = InfrastructureMicroService(self.service)
        connection_details = infra_service.get_service_connection_details(service_name)
        print(connection_details)
        if self.is_successful(connection_details):
            body = connection_details['body']
            return f'http://{body["host"]}:{body["port"]}'
        else:
            raise ValueError(f'Unable to get the base url for service ({service_name})')

    def post_request(self, endpoint, payload):
        _url = self.url + endpoint
        arguments = {
            'json': payload,
            'headers': {
                'Authorization': 'Bearer ' + self.jwt
            }
        }

        response = None
        success = False

        for i in range(5):
            try:
                response = requests.post(_url, **arguments)
                response = response.json()
                success = True
            except:
                time.sleep(0.5)

        if not success:
            response = {
                'header': {
                    'status_code': 500,
                    'code': 'err_api_internal_error',
                    'message': 'Unable to connect to the service after 5 retries.',
                    'body_type': 'failure_info'
                },
                'body': None
            }

        return response

    def is_successful(self, response):
        if response is None:
            return False

        header = response.get('header')
        if header is None:
            return False

        http_code = header.get('status_code')
        if http_code is None or not isinstance(http_code, int):
            return False

        return 200 <= http_code < 300



from abc import ABC, abstractmethod

from convergence.internal.observability.request_log import RequestLog


class ILogSerializer(ABC):
    @abstractmethod
    def save(self, request_log: RequestLog):
        pass


class LogSerializerConfig:
    @staticmethod
    def create(service) -> ILogSerializer:
        from convergence.internal.observability.serializers.json_serializer import JsonLogSerializer
        path = service.get_configuration('observability.path')
        production = service.get_configuration('application.mode') == 'production'
        return JsonLogSerializer(path, production)



from convergence.db_migrations.base.base_table_formatter import BaseTableBlueprintFormatter
from convergence.db_migrations.structures import TableBlueprint, TableIndexBlueprint


class PostgresTableFormatter(BaseTableBlueprintFormatter):
    def format_table_index(self, blueprint: TableBlueprint, index_blueprint: TableIndexBlueprint):
        table = blueprint.name
        fields_comma = []
        fields_underscore = []

        sc = ""
        su = ""

        for col in index_blueprint.columns:
            fields_comma.append(sc)
            fields_comma.append(col)
            fields_underscore.append(su)
            fields_underscore.append(col)
            sc = ", "
            su = "_"

        fields_underscore = ''.join(fields_underscore)
        fields_comma = ''.join(fields_comma)

        return f"CREATE INDEX CONCURRENTLY {table}_{fields_underscore}_index ON {table} USING {index_blueprint.type}({fields_comma})"



from starlette.middleware.base import BaseHTTPMiddleware
from fastapi import Request
from starlette.responses import JSONResponse

from convergence.dto.bad_request_details import BadRequestDetails, BadFieldRequest
from convergence.dto.base_dto import ApiResponse
from convergence.helpers import errors
from convergence.helpers.controller_helpers import convert_object_to_dict, update_response_with_request_ids_for_filters
from convergence.internal.observability.request_log import RequestLog
from convergence.internal.observability.serializer_config import LogSerializerConfig
from convergence.constants import (HEADER_REQUEST_ID,
                                   HEADER_PARENT_REQUEST_ID,
                                   HEADER_CALLER_SERVICE,
                                   HEADER_CALLER_SERVICE_VERSION,
                                   HEADER_CALLER_SERVICE_HASH)


RESERVED_MANDATORY_GATEWAY_HEADERS = [
    HEADER_REQUEST_ID,
    HEADER_CALLER_SERVICE,
    HEADER_CALLER_SERVICE_HASH,
    HEADER_CALLER_SERVICE_VERSION,
]

RESERVED_OPTIONAL_GATEWAY_HEADERS = [
    HEADER_PARENT_REQUEST_ID
]


class GatewayHeaderValidationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, **kwargs):
        super().__init__(app)
        self.service = kwargs['service']
        self.is_behind_gateway = self.service.get_configuration('security.is_behind_gateway')
        self.log_serializer = LogSerializerConfig.create(self.service)

    async def dispatch(self, request: Request, call_next):
        if self.is_behind_gateway:
            # Behind gateway, no need to validate token
            is_missing, missing_headers = self.__is_missing_any_of_gateway_headers(request)
            if is_missing:
                response = self.__request_is_missing_gateway_headers(request, missing_headers)
            else:
                response = await call_next(request)
        elif len(self.__get_independent_service_request_invalid_headers(request)) > 0:
            # Independent service, but contains reserved headers, must fail
            response = self.__request_has_reserved_headers_response(request)
        elif self.__request_has_authorization_header(request):
            # Independent service, has authorization, so must be validated
            if self.__is_authorization_header_valid(request):
                response = await call_next(request)
            else:
                response = self.__unauthorized_response(request)
        else:
            # Independent service, no authorization, can proceed. If the endpoint need authorization, it will be
            # dropped by authorization filter
            response = await call_next(request)

        return response

    def __is_missing_any_of_gateway_headers(self, request):
        headers = request.headers.keys()
        headers = [header.upper() for header in headers]

        missing_headers = []
        for header in RESERVED_MANDATORY_GATEWAY_HEADERS:
            if header not in headers:
                missing_headers.append(header)

        return len(missing_headers) > 0, missing_headers

    def __request_is_missing_gateway_headers(self, http_request, missing_headers):
        request_type = self.service.get_configuration('observability.request_id_prefix')
        request_log = RequestLog.initialize(request_type, http_request, self.service)

        status_code = 502

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = 'The request input was invalid, missing the mandatory API gateway headers.'
        response.header.code = errors.INVALID_DATA
        response.body = BadRequestDetails()
        response.header.body_type = response.body.get_response_body_type()

        for header in missing_headers:
            details = BadFieldRequest()
            details.location = 'header'
            details.field = header
            details.error_messages = ['Request is missing header.']
            response.body.errors.append(details)

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    def __get_independent_service_request_invalid_headers(self, request):
        headers = request.headers.keys()
        headers = [header.upper() for header in headers]

        invalid_headers = []
        for header in RESERVED_MANDATORY_GATEWAY_HEADERS:
            if header in headers:
                invalid_headers.append(header)

        return invalid_headers

    def __request_has_reserved_headers_response(self, request):
        request_type = self.service.get_configuration('observability.request_id_prefix')
        request_log = RequestLog.initialize(request_type, request, self.service)

        invalid_headers = self.__get_independent_service_request_invalid_headers(request)

        status_code = 400

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = 'The request input was invalid, including reserved API gateway headers.'
        response.header.code = errors.INVALID_DATA
        response.body = BadRequestDetails()
        response.header.body_type = response.body.get_response_body_type()

        for header in invalid_headers:
            details = BadFieldRequest()
            details.location = 'header'
            details.field = header
            details.error_messages = ['Request includes reserved header.']
            response.body.errors.append(details)

        update_response_with_request_ids_for_filters(request_log, response, request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    def __request_has_authorization_header(self, request):
        headers = request.headers.keys()
        headers = [header.lower() for header in headers]

        return 'authorization' in headers

    def __is_authorization_header_valid(self, request):
        # TODO: implement
        return True

    def __unauthorized_response(self, http_request):
        request_type = self.service.get_configuration('observability.request_id_prefix')
        request_log = RequestLog.initialize(request_type, http_request, self.service)

        status_code = 403

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = 'Unable to verify the validity of the Authorization token provided.'
        response.header.code = errors.INVALID_AUTHORIZATION_TOKEN
        response.header.body_type = 'failure_info'

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)



VERSION = '0.0.6'
VERSION_HASH = ''
BUILD_DATE = '2023-11-19 11:57:03 UTC'

HEADER_REQUEST_ID = "X-CONVERGENCE-REQUEST-ID"
HEADER_CALLER_SERVICE = "X-CONVERGENCE-CALLER-SERVICE"
HEADER_CALLER_SERVICE_HASH = "X-CONVERGENCE-CALLER-SERVICE-HASH"
HEADER_CALLER_SERVICE_VERSION = "X-CONVERGENCE-CALLER-SERVICE-VERSION"
HEADER_PARENT_REQUEST_ID = "X-CONVERGENCE-PARENT-REQUEST-ID"
